////////////////////////////////////////////////////////////////////////////////
// decode.cpp
//   An example tool for reconstructing the text from a file containing the LZ77
//   parsing.
////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2013 Juha Karkkainen, Dominik Kempa and Simon J. Puglisi
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use,
// copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following
// conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <fstream>

#include <cstdlib>
#include <ctime>

#include "common.h"

int main(int argc, char **argv) {
  if (argc != 3) {
    std::cerr << "usage: " << argv[0] << " infile outfile\n\n"
      << "Reconstructs the text from infile, that contains \n"
      << "LZ77 parsing and writes the output to outfile.   \n";
    exit(EXIT_FAILURE);
  }

  // Open the file with parsing.
  std::cerr << "Opening the file " << argv[1] << "... ";
  std::fstream f(argv[1], std::fstream::in);
  if (f.fail()) {
    std::cerr << "\nError: cannot open file " << argv[1] << "\n";
    std::exit(EXIT_FAILURE);
  }
  std::cerr << std::endl;

  // Read the text length.
  int length;
  f.read((char *)&length, sizeof(int));
  unsigned char *text = new unsigned char[length];
  if (!text) {
    std::cerr << "\nError: allocation of " << length << " bytes failed\n";
    std::exit(EXIT_FAILURE);
  }

  // Reconstruct the text from the parsing.
  std::cerr << "Decoding the parsing... ";
  std::clock_t timestamp = std::clock();
  int restored = 0, phrase[2];
  while (f.read((char *)phrase, 2 * sizeof(int))) {
    int pos = phrase[0], len = phrase[1];
    if (restored == length) {
      std::cerr << "\nError: restored string is too long (" << restored << ")\n";
      std::exit(EXIT_FAILURE);
    }
    if (!len) text[restored++] = pos;
    else {
      if (pos < 0 || pos >= restored) {
        std::cerr << "\nError: nonsense pos value (" << pos << ") after byte "
          << restored << "\n";
        std::exit(EXIT_FAILURE);
      } else if (restored + len > length) {
        std::cerr << "\nError: too big len value (" << len << ") after byte "
          << restored << "\n";
        std::exit(EXIT_FAILURE);
      } else {
        for (int j = 0; j < len; ++j)
          text[restored++] = text[pos + j];
      }
    }
  }
  std::cerr << elapsed(timestamp) << "s\n";
  f.close();
  if (restored != length) {
    std::cerr << "\nError: restored text is too short (" << restored << ")\n";
    std::exit(EXIT_FAILURE);
  }

  // Write the output to file.
  std::cerr << "Writing the text to " << argv[2] << "... ";
  std::fstream outfile(argv[2], std::fstream::out);
  outfile.write((char *)text, length);
  outfile.close();
  std::cerr << std::endl;

  // Clean up.
  delete[] text;
  return EXIT_SUCCESS;
}
