////////////////////////////////////////////////////////////////////////////////
// encode.cpp
//   An example tool computing the LZ77 parsing of a given file.
////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2013 Juha Karkkainen, Dominik Kempa and Simon J. Puglisi
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use,
// copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following
// conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <fstream>
#include <vector>

#include <ctime>

#include "../algorithm/lzisa.h"
#include "common.h"

int main(int argc, char **argv) {
  // Check arguments.
  if(argc != 3 && argc != 4) {
    std::cerr << "usage: " << argv[0] << " infile outfile [algorithm]\n\n"
      << "Computes the LZ77 parsing of infile using selected           \n"
      << "algorithm (optional) and writes the result to outfile.       \n"
      << "Available algorithms are:                                    \n"
      << "  lzisa9  -- the fastest, uses 9n bytes (default)            \n"
      << "  lzisa6s -- slower than lzisa9 but uses only 6n bytes       \n"
      << "  lzisa6r -- fast 6n bytes version for repetitive data       \n";
    std::exit(EXIT_FAILURE);
  }

  // Read the text and its suffix array.
  unsigned char *text;
  int *sa, length;
  read_text(argv[1], text, length);
  read_sa(argv[1], sa, length);

  // Compute the parsing.
  std::clock_t timestamp = std::clock();
  std::string alg = (argc == 3) ? "lzisa9" : argv[3];
  int nphrases;
  std::vector<std::pair<int, int> > phrases;
  if (alg == "lzisa9") {
    std::cerr << "Running algorithm lzisa9... ";
    nphrases = lzisa9(text, sa, length, &phrases);
  } else if (alg == "lzisa6s") {
    std::cerr << "Running algorithm lzisa6s... ";
    nphrases = lzisa6s(text, sa, length, &phrases);
  } else if (alg == "lzisa6r") {
    std::cerr << "Running algorithm lzisa6r... ";
    nphrases = lzisa6r(text, sa, length, &phrases);
  } else {
    std::cerr << "\nError: unrecognized algorithm name\n";
    std::exit(EXIT_FAILURE);
  }
  std::cerr << elapsed(timestamp) << "s\n";
  
  // Write the parsing to file.
  std::cerr << "Writing " << nphrases << " phrases to " << argv[2] << "... ";
  std::fstream outfile(argv[2], std::fstream::out);
  outfile.write((char *)&length, sizeof(int));
  for (int i = 0; i < (int)phrases.size(); ++i) {
    outfile.write((char *)&phrases[i].first, sizeof(int));
    outfile.write((char *)&phrases[i].second, sizeof(int));
  }
  outfile.close();
  std::cerr << std::endl;

  // Clean up.
  delete[] sa;
  delete[] text;
  return EXIT_SUCCESS;
}
